import openpyxl
from viggocore.common.subsystem import operation
from viggocore.subsystem.capability.resource import Capability
from viggocore.subsystem.role.resource import Role
from viggocore.subsystem.route.resource import Route
from sqlalchemy import func


class ImportCapabilitiesAndPolicies(operation.Operation):
    """
    Importa uma lista de fabricantes via planilha xlsx
    """

    # função genérica para pegar os dados da planilha
    def _get_sheet_data(self, sheet) -> list:
        """
        Lê os dados da planilha e os converte em uma lista de dicionários

        :param sheet: tabela de dados
        :return: lista de dicionários
        :rtype: list
        """

        # cria lista de retorno
        result = []
        # seta linha inicial como 2, a primeira linha é dos nomes das colunas
        row = 2

        # itera sobre os dados da tabela até que a primeira coluna da linha
        # atual esteja vazia, significando o final da tabela
        while sheet.cell(row=row, column=1).value is not None:
            # define dicionário auxiliar para objeto sendo iterado
            aux = {}
            # inicia na coluna 1
            col = 1

            # itera sobre as colunas da linha atual até que o valor observado
            # seja nulo
            while sheet.cell(row=1, column=col).value is not None:
                # atualiza dicionário auxiliar com valor observado na iteração
                # e chave sendo o nome da coluna
                aux.update({
                    sheet.cell(row=1, column=col).value:
                    sheet.cell(row=row, column=col).value
                })
                col += 1

            # appende dicionário auxiliar à lista de retorno
            result.append(aux)
            row += 1

        return result

    def _get_route_ids(self, session, sheet_data):
        filtro = [
            f'{data.get("URL", None)} - {data.get("METHOD", None)}'
            for data in sheet_data]
        query = session.query(Route.id)\
            .filter(func.concat(Route.url, ' - ', Route.method).in_(filtro))
        result = query.all()
        result = [r[0] for r in result]
        return result

    def _get_roles(self, session, sheet_data):
        role_names = set([data.get("PAPEL", 'NULL').upper()
                          for data in sheet_data])
        query = session.query(Role)\
            .filter(func.upper(Role.name).in_(role_names))
        result = query.all()
        return {r.name: r.id for r in result}

    def _get_role_id_cabilities(self, session, sheet_data):
        role_id_cabilities = {}
        roles = self._get_roles(session, sheet_data)
        for role_name, role_id in roles.items():
            sheet_data_filtred = list(
                filter(lambda x: x.get('PAPEL', '') == role_name, sheet_data))
            filtro = [
                f'{data.get("URL", None)} - {data.get("METHOD", None)}'
                for data in sheet_data_filtred]
            query = session.query(Capability.id)\
                .join(Route, Route.id == Capability.route_id)\
                .filter(Capability.application_id == self.id)\
                .filter(func.concat(Route.url, ' - ', Route.method).in_(filtro))
            result = query.all()
            role_id_cabilities[role_id] = [r[0] for r in result]

        return role_id_cabilities

    # função que monta os dicionários das capacidades para serem salvas
    def _map_capabilities_dict(self, session, sheet_data):
        capabilities_dict = []
        route_ids = self._get_route_ids(session=session, sheet_data=sheet_data)
        for route_id in route_ids:
            capabilities_dict.append({
                "application_id": self.id,
                "route_id": route_id
            })
        return capabilities_dict

    # função que salva as capacidades
    def _save_capabilities(self, session, wb_obj):
        # valida se a planilha possui a tabela apropriada para a importação
        # dados e lança uma BadRequest caso a tabela não seja encontrada
        sheet_obj = list(filter(
            lambda x: x.title == 'CAPACIDADES', wb_obj.worksheets))
        if len(sheet_obj) == 0:
            return

        # converte tabela de objetos em lista de dicionários
        sheet_obj = sheet_obj[0]
        sheet_data = self._get_sheet_data(sheet_obj)
        # mapeia a lista de dicionários em objetos na estrutura do DB
        capabilities_dict = self._map_capabilities_dict(session, sheet_data)

        # salva objetos e suas dependências no DB
        for cap_dict in capabilities_dict:
            try:
                self.manager.api.capabilities().create(
                    session=session, **cap_dict)
                session.commit()
                # self.driver.update(session=session)
                self.capabilities_sucessos += 1
            except Exception as e:
                print(str(e))
                self.capabilities_falhas += 1
                session.rollback()

    # função que monta os dicionários das capacidades para serem salvas
    def _map_policies_dict(self, session, sheet_data):
        policies_dict = []
        role_cabilities = self._get_role_id_cabilities(session, sheet_data)
        for role_id in role_cabilities.keys():
            cability_ids = role_cabilities.get(role_id, [])
            for c_id in cability_ids:
                policies_dict.append({
                    "role_id": role_id,
                    "capability_id": c_id
                })
        return policies_dict

    # função que salva as politicas
    def _save_policies(self, session, wb_obj):
        # valida se a planilha possui a tabela apropriada para a importação
        # dados e lança uma BadRequest caso a tabela não seja encontrada
        sheet_obj = list(filter(
            lambda x: x.title == 'POLITICAS', wb_obj.worksheets))
        if len(sheet_obj) == 0:
            return

        # converte tabela de objetos em lista de dicionários
        sheet_obj = sheet_obj[0]
        sheet_data = self._get_sheet_data(sheet_obj)
        # mapeia a lista de dicionários em objetos na estrutura do DB
        policies_dict = self._map_policies_dict(session, sheet_data)

        # salva objetos e suas dependências no DB
        for policy_dict in policies_dict:
            try:
                self.manager.api.policies().create(
                    session=session, **policy_dict)
                self.policies_sucessos += 1
                session.commit()
            except Exception as e:
                print(str(e))
                self.policies_falhas += 1
                session.rollback()

    def pre(self, id, **kwargs):
        self.id = id
        # recupera a planilha à partir do dicionário kwargs
        self.file = kwargs.get('file', None)
        # recupera domain_id à partir do dicionário kwargs
        self.domain_id = kwargs.get('domain_id', None)

        return True

    def do(self, session, **kwargs):
        self.capabilities_sucessos = 0
        self.capabilities_falhas = 0
        self.policies_sucessos = 0
        self.policies_falhas = 0

        # inicia o arquivo xlsx
        wb_obj = openpyxl.load_workbook(self.file)

        self._save_capabilities(session=session, wb_obj=wb_obj)
        self._save_policies(session=session, wb_obj=wb_obj)

        print(f"""
        capabilities_sucessos: {self.capabilities_sucessos}
        capabilities_falhas: {self.capabilities_falhas}
        policies_sucessos: {self.policies_sucessos}
        policies_falhas: {self.policies_falhas}
        """)

        return super().do(**kwargs)



QUERY = """
DO $$ 
DECLARE
    v_route_id CHAR(32);
    v_application_id CHAR(32);
    v_application_name TEXT := '{application_name}';
BEGIN
    -- Obtenha application_id uma vez, pois é constante
    SELECT id INTO v_application_id
    FROM "application"
    WHERE "name" = v_application_name;

    FOR v_route_id IN 
        SELECT r.id
        FROM route r
        WHERE concat(r.url, ' - ', r."method") IN (
            ''
            {url_methods}
        )
    LOOP
        BEGIN
            -- Insira na tabela capability
            INSERT INTO capability (id, active, created_at, tag, route_id, application_id)
            VALUES (md5((random())::text), TRUE, now(), '#INSERIDO_VIA_SCRIPT', v_route_id, v_application_id);

        EXCEPTION WHEN OTHERS THEN
            -- Ignore o erro e continue
            RAISE NOTICE 'Erro ao inserir route_id %: % na capability', v_route_id, SQLERRM;
        END;
    END LOOP;
END;
$$ LANGUAGE plpgsql;
"""  # noqa

QUERY_FILTER = ", '{url} - {method}'"  # noqa



from viggocore.common import utils as core_utils
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl import Workbook, styles
from viggocore.common.subsystem.manager_functions import make_exportar_xlsx


class GetXlsxImportModel(make_exportar_xlsx.MakeExportarXsls):

    # funções sobrescritas
    def _formatar_celulas(self, ws, dados, cols, is_to_bold=True):
        qtd_linhas = len(dados)

        if qtd_linhas > 0:
            lin_ini = 2
            lin_fim = lin_ini + qtd_linhas  # A última linha é de total

            for lin in range(lin_ini, lin_fim + 1):
                # Accounting format
                for col in cols:
                    cell = ws.cell(lin, col['num'])
                    cell.border = styles.Border(
                        top=styles.Side(border_style='thin'),
                        right=styles.Side(border_style='thin'),
                        bottom=styles.Side(border_style='thin'),
                        left=styles.Side(border_style='thin'))

                    if 'fmt' in col:
                        cell.number_format = col['fmt']

                    if 'align' in col:
                        cell.alignment = styles.Alignment(
                            horizontal=col['align'])

                    if 'function' in col:
                        formula = col['function'] \
                            .replace('[row]', str(cell.row))
                        cell.value = formula

                    if (lin == lin_fim):
                        self._formatar_footer(
                            col, cell, lin_ini, qtd_linhas)
        return

    # função para pegar o nome do arquivo
    def _get_file_name(self, **kwargs):
        return 'import-model-capabilities-and-policies.xlsx'

    # funções de preenchimento do xlsx
    def _gen_work_sheet_capabilities(self, ws: Worksheet, capabilities):
        ROW_INIT = 1

        cols = [
            {'num': 1, 'name': 'URL', 'width': 50},
            {'num': 2, 'name': 'METHOD', 'width': 25}
        ]

        self._def_header_style(ws, cols, 'export_capabilities',
                               row_ini=ROW_INIT)

        for linha in capabilities:
            ws.append(list(linha))

        self._formatar_celulas(ws, capabilities, cols)
        self._style_header(ws, cols)
        self._align_horizontal_center_column(ws, capabilities, 2)

        return True

    def _gen_work_sheet_policies(self, ws: Worksheet, policies):
        ROW_INIT = 1

        cols = [
            {'num': 1, 'name': 'PAPEL', 'width': 25},
            {'num': 2, 'name': 'URL', 'width': 50},
            {'num': 3, 'name': 'METHOD', 'width': 25}
        ]

        self._def_header_style(ws, cols, 'export_policies',
                               row_ini=ROW_INIT)

        for linha in policies:
            ws.append(list(linha))

        self._formatar_celulas(ws, policies, cols)
        self._style_header(ws, cols)
        self._align_horizontal_center_column(ws, policies, 2)

        return True

    def pre(self, **kwargs):
        return True

    def do(self, session, **kwargs):
        capabilities = []
        policies = []

        self.file_name = self._get_file_name(**kwargs)
        self.download_folder = core_utils.get_upload_folder()

        wb = Workbook()

        # Aba de capacidades
        ws_cap = wb.active  # Usa a aba ativa pra primeira
        ws_cap.title = "CAPACIDADES"
        self._gen_work_sheet_capabilities(ws_cap, capabilities)

        # Aba de políticas
        ws_pol = wb.create_sheet(title="POLITICAS")
        self._gen_work_sheet_policies(ws_pol, policies)

        wb.save(filename=self.download_folder + '/' + self.file_name)

        return self.download_folder, self.file_name



from viggocore.common import exception
from viggocore.common import utils as core_utils
from viggocore.subsystem.capability.resource import Capability
from viggocore.subsystem.role.resource import Role
from viggocore.subsystem.route.resource import Route
from viggocore.subsystem.policy.resource import Policy
from sqlalchemy import func, or_
from viggocore.common.subsystem import operation
from viggocore.subsystem.application.functions.template import (
    insert_capabilities_sql as ic)
from viggocore.subsystem.application.functions.template import (
    insert_policies_sql as ip)
from datetime import datetime as datetime1


class ExportCapabilitiesAndPoliciesSql(operation.List):

    def _montar_sql_capabilities(self, capabilities):
        url_methods = ''
        for c in capabilities:
            url_methods += ic.QUERY_FILTER.format(url=c[0], method=c[1])

        return ic.QUERY.format(
            application_name=self.application_name,
            url_methods=url_methods)

    def _montar_sql_policies(self, policies):
        sql_policies = ''
        roles = list(set([p[0] for p in policies]))
        for role in roles:
            url_methods = ''
            policies_filtred = list(filter(lambda x: x[0] == role, policies))
            for policy in policies_filtred:
                url_methods += ip.QUERY_FILTER.format(
                    url=policy[1], method=policy[2])

            sql_policies += ip.QUERY.format(
                role_name=role,
                application_name=self.application_name,
                url_methods=url_methods)

        return sql_policies

    def _montar_sql(self, capabilities, policies):
        conteudo_sql = ''
        conteudo_sql += self._montar_sql_capabilities(capabilities)
        conteudo_sql += self._montar_sql_policies(policies)
        return conteudo_sql

    # função para pegar o nome do arquivo
    def _get_file_name(self, **kwargs):
        application = self.manager.get(id=self.id)
        self.application_name = application.name

        file_name = kwargs.get('file_name', None)
        if file_name is not None:
            return file_name
        else:
            file_name = f'{self.application_name}-{self.a_partir}.sql'
            return file_name

    # valida se o tipo de exportacao passado é válido
    def _validar_tipo_exportacao(self):
        tipos_permitidos = ['CAPACIDADE', 'POLITICA']
        if any(list(map(
                lambda x: x not in tipos_permitidos, self.tipo_exportacao))):
            raise exception.BadRequest(
                'Os tipos de exportação permitidos são:' +
                ', '.join(tipos_permitidos) + '.')

    # funções de buscar dados
    def _get_capabilities(self, session):
        query = session.query(Route.url, Route.method)\
            .join(Capability, Capability.route_id == Route.id)\
            .filter(Capability.application_id == self.id)\
            .filter(or_(Route.created_at >= self.a_partir,
                        Route.created_at.is_(None)))\
            .order_by(Route.url, Route.method)
        result = query.all()
        return result

    def _get_policies(self, session):
        papeis = [p.upper() for p in self.papeis]
        query = session.query(Role.name, Route.url, Route.method)\
            .join(Policy, Policy.role_id == Role.id)\
            .join(Capability, Capability.id == Policy.capability_id)\
            .join(Route, Route.id == Capability.route_id)\
            .filter(func.upper(Role.name).in_(papeis))\
            .filter(Capability.application_id == self.id)\
            .filter(or_(Route.created_at >= self.a_partir,
                        Route.created_at.is_(None)))\
            .order_by(Role.name, Route.url, Route.method)
        result = query.all()
        return result

    def _get_capabilities_and_policies(self, session):
        # inicia as listas de capacidades e políticas
        capabilities = []
        policies = []

        # se foi passada a CAPACIDADE como tipo_exportacao então vai buscar as
        # capacidades
        if 'CAPACIDADE' in self.tipo_exportacao:
            capabilities = self._get_capabilities(session=session)

        # se foi passada a POLITICA como tipo_exportacao então vai buscar as
        # políticas ligadas as capacidades desta aplicação
        if 'POLITICA' in self.tipo_exportacao:
            policies = self._get_policies(session=session)

        return (capabilities, policies)

    def pre(self, id, **kwargs):
        self.id = id
        if self.id is None:
            raise exception.BadRequest(
                'É obrigatório passar o id na rota da requisição.')

        self.tipo_exportacao = kwargs.pop('tipo_exportacao', [])
        if not self.tipo_exportacao:
            raise exception.BadRequest(
                'É obrigatório informar o tipo_exportacao.')

        self._validar_tipo_exportacao()

        self.papeis = kwargs.pop('papeis', [])
        if 'POLITICA' in self.tipo_exportacao and not self.papeis:
            raise exception.BadRequest(
                'É obrigatório informar os papéis caso você ' +
                'queira exportar as políticas também.')

        self.a_partir = kwargs.pop('a_partir', None)
        if self.a_partir is None:
            raise exception.BadRequest(
                'É obrigatório informar o campo "a_partir" para filtrar ' +
                'as informações que foram geradas posterior a esta data.')
        else:
            try:
                self.a_partir = datetime1.strptime(
                    self.a_partir.replace(' ', '+'), '%Y-%m-%d%z')
            except Exception:
                raise exception.BadRequest(
                    'É obrigatório passar a data no formato "yyyy-mm-dd-tz."')

        return True

    def do(self, session, **kwargs):
        capabilities, policies = self._get_capabilities_and_policies(
            session=session)

        if len(capabilities) == 0 and len(policies) == 0:
            raise exception.BadRequest(
                'Não existe dados para essa filtragem.')

        self.file_name = self._get_file_name(**kwargs)
        self.download_folder = core_utils.get_upload_folder()
        path = self.download_folder + '/' + self.file_name

        conteudo_sql = self._montar_sql(capabilities, policies)

        # salva o arquivo sql
        with open(path, "w", encoding="utf-8") as arquivo:
            arquivo.write(conteudo_sql)

        return self.download_folder, self.file_name



QUERY = """
DO $$
DECLARE
    v_id UUID;
    v_capability_id CHAR(32);
    v_role_id CHAR(32);
    v_application_name TEXT := '{application_name}';
BEGIN
    FOR v_capability_id IN
        SELECT c.id
        FROM capability c
        JOIN "route" r ON r.id = c.route_id
        JOIN "application" a ON a.id = c.application_id
        WHERE a."name" = v_application_name
          AND concat(r.url, ' - ', r."method") IN (
              ''
              {url_methods}
          )
    LOOP
        BEGIN
            -- Obtenha role_id
            SELECT id INTO v_role_id
            FROM "role"
            WHERE upper("name") = upper('{role_name}');

            -- Insira na tabela policy
            INSERT INTO public."policy" (id, active, created_at, tag, capability_id, role_id)
            VALUES (md5((random())::text), TRUE, now(), '#INSERIDO_VIA_SCRIPT', v_capability_id, v_role_id);

        EXCEPTION WHEN OTHERS THEN
            -- Ignore o erro e continue
            RAISE NOTICE 'Erro ao inserir capability_id na policy%: %', v_capability_id, SQLERRM;
        END;
    END LOOP;
END;
$$ LANGUAGE plpgsql;

"""  # noqa

QUERY_FILTER = ", '{url} - {method}'"  # noqa





import flask

from viggocore.common import exception, utils


def get_xlsx_import_model(self):
    try:
        download_folder, file_name = self.manager.get_xlsx_import_model()
    except exception.ViggoCoreException as exc:
        return flask.Response(response=exc.message,
                              status=exc.status)

    response = flask.send_from_directory(download_folder, file_name)
    utils.remove_file(f'{download_folder}/{file_name}')
    return response



import flask
from viggocore.common import exception


def import_capabilities_and_policies(self, id):
    try:
        # recupera o arquivo à partir da requisição flask
        file = flask.request.files.get('file', None)

        # valida se arquivo foi enviado e envia BadRequest
        if not file:
            raise exception.BadRequest(
                'ERRO! Arquivo não enviado na requisição')

        # adiciona arquivo ao dicionário
        data = {'file': file}

        self.manager.import_capabilities_and_policies(id=id, **data)
    except exception.ViggoCoreException as exc:
        # retorna uma exceção no caso de falha
        return flask.Response(response=exc.message,
                              status=exc.status)

    # retorna um status 204 no sucesso
    return flask.Response(status=204,
                          mimetype=self.MIMETYPE_JSON)
